package com.sd.business.service.work.impl;

import cn.hutool.core.util.ArrayUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.google.common.collect.HashBasedTable;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.exception.ServiceException;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.in.dto.InOutStorageDto;
import com.sd.business.entity.in.emums.InDetailTypeEnum;
import com.sd.business.entity.in.emums.InOutTypeEnum;
import com.sd.business.entity.in.emums.OutDetailTypeEnum;
import com.sd.business.entity.in.po.InOutStorageBom;
import com.sd.business.entity.order.enums.OrderExceptionTypeEnum;
import com.sd.business.entity.order.enums.OrderStatusEnum;
import com.sd.business.entity.order.po.OrderInfo;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.order.po.OrderSkuBom;
import com.sd.business.entity.sku.po.SkuSpec;
import com.sd.business.entity.warehouse.constant.WarehouseConstant;
import com.sd.business.entity.work.constant.MaterialsConstant;
import com.sd.business.entity.work.dto.WorkOrderDto;
import com.sd.business.entity.work.dto.WorkOrderSelectDto;
import com.sd.business.entity.work.enums.WorkOrderFixationSpecEnum;
import com.sd.business.entity.work.po.WorkOrder;
import com.sd.business.entity.work.po.WorkOrderDetail;
import com.sd.business.entity.work.vo.OpenSeaWorkOrderVo;
import com.sd.business.entity.work.vo.WorkOrderDetailVo;
import com.sd.business.entity.work.vo.WorkOrderVo;
import com.sd.business.mapper.work.WorkOrderMapper;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.in.InOutStorageService;
import com.sd.business.service.order.OrderInfoService;
import com.sd.business.service.order.OrderSkuBomService;
import com.sd.business.service.order.OrderSkuService;
import com.sd.business.service.sku.SkuSpecService;
import com.sd.business.service.work.WorkOrderDetailService;
import com.sd.business.service.work.WorkOrderService;
import com.sd.business.util.CodeEnum;
import com.sd.business.util.fixation.CoordinateBo;
import com.sd.business.util.fixation.FixationFactory;
import com.sd.business.util.fixation.FixationStrategy;
import com.sd.business.util.packing.GA;
import com.sd.business.util.packing.entity.PlaceItem;
import com.sd.business.util.packing.entity.Solution;
import com.sd.business.util.packing.model.Item;
import com.sd.framework.util.Assert;
import com.sd.framework.util.TransactionUtil;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;


/**
 * <p>
 * 生产工单 服务实现类
 * </p>
 *
 * @author
 * @since 2024-01-02
 */
@Service
public class WorkOrderServiceImpl extends ServiceImpl<WorkOrderMapper, WorkOrder> implements WorkOrderService {

    public static final ReentrantLock schedulingLock = new ReentrantLock();

    @Autowired
    private BomSpecService bomSpecService;

    @Autowired
    private SkuSpecService skuSpecService;

    @Autowired
    private OrderInfoService orderInfoService;

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private WorkOrderDetailService workOrderDetailService;

    @Autowired
    private InOutStorageService inStorageService;

    @Autowired
    private OrderSkuBomService orderSkuBomService;

    @Override
    public Page<WorkOrderVo> getPage(WorkOrderSelectDto dto) {

        Page<WorkOrderVo> page = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)

                .from(WorkOrder.class)

                .orderByDesc(WorkOrder::getId)
                .eq(WorkOrder::getSyncProduction, dto.getSyncProduction())
                .eq(WorkOrder::getMasterBomSpecId, dto.getMasterBomSpecId())
                .eq(WorkOrder::getPrintingPaperBomSpecId, dto.getPrintingPaperBomSpecId())
                .between(WorkOrder::getCreateTime, dto.getBeginTime(), dto.getEndTime())
                .page(dto);

        List<WorkOrderVo> records = page.getRecords();
        if (records.isEmpty()) {
            return page;
        }

        Set<Long> bomSpecIdSet = records.stream()
                .flatMap(item -> Stream.of(item.getMasterBomSpecId(), item.getPrintingPaperBomSpecId()))
                .collect(Collectors.toSet());

        Map<Long, BomSpec> map = bomSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, bomSpecIdSet));

        for (WorkOrderVo vo : records) {
            BomSpec masterBomSpec = map.get(vo.getMasterBomSpecId());
            if (masterBomSpec != null) {
                vo.setMasterBomSpecCode(masterBomSpec.getCode());
                vo.setMasterBomSpecName(masterBomSpec.getName());
            }

            BomSpec printingPaperBomSpec = map.get(vo.getPrintingPaperBomSpecId());
            if (printingPaperBomSpec != null) {
                vo.setPrintingPaperBomSpecCode(printingPaperBomSpec.getCode());
                vo.setPrintingPaperBomSpecName(printingPaperBomSpec.getName());
            }
        }

        return page;
    }

    @Override
    public WorkOrderVo detail(Long id) {

        WorkOrderVo vo = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)
                .from(WorkOrder.class)
                .eq(WorkOrder::getId, id)
                .one();

        Assert.notNull(vo, "未知数据");

        // 排版信息
        List<WorkOrderDetailVo> workOrderDetailList = Sql.create(WorkOrderDetailVo.class)
                .selectAll(WorkOrderDetail.class)
                .selectAs(OrderInfo::getCode, WorkOrderDetailVo::getOrderCode)
                .selectAs(OrderInfo::getWlnCode, WorkOrderDetailVo::getOrderWlnCode)
                .selectAs(SkuSpec::getCode, WorkOrderDetailVo::getSkuSpecCode)
                .selectAs(SkuSpec::getName, WorkOrderDetailVo::getSkuSpecName)

                .from(WorkOrderDetail.class)
                .leftJoin(OrderInfo.class, WorkOrderDetail::getOrderInfoId)
                .leftJoin(OrderSku.class, WorkOrderDetail::getOrderSkuId)
                .leftJoin(SkuSpec.class, OrderSku::getSkuSpecId)

                .eq(WorkOrderDetail::getWorkOrderId, id)
                .list();

        vo.setWorkOrderDetailList(workOrderDetailList);

        // 公海订单
        List<OrderSku> orderSkuList = getNotSchedulingSku(vo.getMasterBomSpecId(), vo.getPrintingPaperBomSpecId(), null, null);
        List<OpenSeaWorkOrderVo> openSeaWorkOrderList = getOpenSeaWorkOrderList(orderSkuList);
        orderInfoService.attributeAssign(openSeaWorkOrderList, OpenSeaWorkOrderVo::getOrderInfoId, (item, orderInfo) -> {
            item.setOrderCode(orderInfo.getCode());
            item.setOrderWlnCode(orderInfo.getWlnCode());
        });
        skuSpecService.attributeAssign(openSeaWorkOrderList, OpenSeaWorkOrderVo::getSkuSpecId, (item, skuSpec) -> {
            item.setSkuSpecCode(skuSpec.getCode());
            item.setSkuSpecName(skuSpec.getName());
            item.setLength(skuSpec.getLength());
            item.setWidth(skuSpec.getWidth());
            item.setBleedingLength(skuSpec.getLength().add(new BigDecimal(MaterialsConstant.RESERVE)));
            item.setBleedingWidth(skuSpec.getWidth().add(new BigDecimal(MaterialsConstant.RESERVE)));
        });
        vo.setOpenSeaWorkOrderList(openSeaWorkOrderList);

        return vo;
    }

    @Override
    public void scheduling() {
        commonScheduling(this::scheduling);
    }

    @Override
    public void schedulingRemaining() {
        commonScheduling((bomSpecId, printingPaperBomSpecId, notFixationList, skuSpecMap) -> {

            WorkOrder workOrder = new WorkOrder();
            workOrder.setId(IdWorker.getId());
            workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
            workOrder.setType(2);
            workOrder.setMasterBomSpecId(bomSpecId);
            workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);

            notFixationScheduling(workOrder, notFixationList, skuSpecMap, false);
        });
    }

    @Override
    public Integer notSchedulingSkuNum() {
        List<OrderSku> orderSkuList = getNotSchedulingSku(null, null, null, null);
        return orderSkuList.stream()
                .mapToInt(item -> item.getQuantity().subtract(item.getSchedulingNum()).intValue())
                .sum();
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void saveScheduling(WorkOrderDto dto) {
        schedulingLock.lock();
        try {
            Long id = dto.getId();
            Assert.notNull(id, "id不能为空");

            List<WorkOrderDetail> oldList = workOrderDetailService.list(q -> q.eq(WorkOrderDetail::getWorkOrderId, id));
            List<WorkOrderDetail> newList = dto.getWorkOrderDetailList();

            // 新增订单skuId集合
            List<Long> addOrderSkuIdList = newList.stream().filter(item -> item.getId() == null)
                    .map(WorkOrderDetail::getOrderSkuId).collect(Collectors.toList());

            // 删除订单skuId集合
            Set<Long> newIdList = newList.stream().map(BaseIdPo::getId).collect(Collectors.toSet());
            List<Long> deleteOrderSkuIdList = oldList.stream().filter(item -> !newIdList.contains(item.getId()))
                    .map(WorkOrderDetail::getOrderSkuId).collect(Collectors.toList());

            Set<Long> orderSkuIdSet = Stream.concat(addOrderSkuIdList.stream(), deleteOrderSkuIdList.stream())
                    .collect(Collectors.toSet());

            Map<Long, OrderSku> map = Collections.emptyMap();
            if (!orderSkuIdSet.isEmpty()) {
                map = orderSkuService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, orderSkuIdSet));
            }

            for (Long item : deleteOrderSkuIdList) {
                OrderSku orderSku = map.get(item);
                Assert.notNull(orderSku, "未知订单skuId：" + item);
                orderSku.setSchedulingNum(orderSku.getSchedulingNum().subtract(BigDecimal.ONE));
                orderSku.setCompleteScheduling(StatusConstant.NO);
                if (orderSku.getSchedulingNum().compareTo(BigDecimal.ZERO) < 0) {
                    throw new ServiceException("订单已排版数量小于0");
                }
            }
            for (Long item : addOrderSkuIdList) {
                OrderSku orderSku = map.get(item);
                Assert.notNull(orderSku, "未知订单skuId：" + item);
                orderSku.setSchedulingNum(orderSku.getSchedulingNum().add(BigDecimal.ONE));
                orderSku.setCompleteScheduling(orderSku.getSchedulingNum().compareTo(orderSku.getQuantity()) == 0
                        ? StatusConstant.YES : StatusConstant.NO);
                if (orderSku.getSchedulingNum().compareTo(orderSku.getQuantity()) > 0) {
                    throw new ServiceException("排版数量超过订单数量");
                }
            }

            updateById(dto);
            workOrderDetailService.editLinked(newList, WorkOrderDetail::getWorkOrderId, id);
            if (!map.values().isEmpty()) {
                orderSkuService.updateBatchById(map.values());
            }

        } finally {
            schedulingLock.unlock();
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void syncScheduling(Long id) {
        WorkOrder workOrder = getById(id);
        Assert.notNull(workOrder, "排版工单不存在");
        Assert.eqZero(workOrder.getSyncProduction(), "排版已同步生产");
        Assert.notNull(workOrder.getMasterInventoryId(), "未选择转印布");
        Assert.notNull(workOrder.getPrintingPaperInventoryId(), "未选择打印纸");

        workOrder.setSyncProduction(StatusConstant.YES);
        updateById(workOrder);

        // 原料出库
        ingredientsOut(workOrder);

        // 余料入库
        surplusIn(workOrder);

        // 获取工单明细
        List<WorkOrderDetail> list = workOrderDetailService.list(q -> q.eq(WorkOrderDetail::getWorkOrderId, workOrder.getId()));

        // 包材出库
        packagingMaterialOut(workOrder, list);

        // 订单状态改为生产中
        Set<Long> orderInfoIdList = list.stream().map(WorkOrderDetail::getOrderInfoId).collect(Collectors.toSet());
        orderInfoService.update(q -> q
                .in(BaseIdPo::getId, orderInfoIdList)
                .eq(OrderInfo::getStatus, OrderStatusEnum.SCHEDULING.getKey())
                .set(OrderInfo::getStatus, OrderStatusEnum.IN_PRODUCTION.getKey())
        );

    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void reschedule(Long id) {

        boolean b = schedulingLock.tryLock();

        if (!b) {
            throw new ServiceException("正在排版中，请稍后刷新数据");
        }

        try {

            WorkOrder workOrder = getById(id);
            if (Objects.equals(workOrder.getSyncProduction(), StatusConstant.YES)) {
                throw new ServiceException("工单已同步生产，无法同步生产");
            }

            List<WorkOrderDetail> workOrderDetailList = workOrderDetailService.list(q -> q.eq(WorkOrderDetail::getWorkOrderId, id));

            // 排版
            reschedule(workOrder, workOrderDetailList);

        } catch (ServiceException e) {
            throw e;
        } catch (Exception e) {
            log.error("排版异常", e);
            throw new ServiceException("排版异常");
        } finally {
            schedulingLock.unlock();
        }
    }

    /**
     * 排版
     */
    private void scheduling(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap) {

        if (ObjectUtil.isEmpty(list)) {
            return;
        }

        // 固定排版订单集合
        List<OrderSku> fixationList = new ArrayList<>();

        // 不固定排版订单集合
        List<OrderSku> notFixationList = new ArrayList<>();

        for (OrderSku orderSku : list) {
            Long skuSpecId = orderSku.getSkuSpecId();
            SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
            boolean fixation = WorkOrderFixationSpecEnum.isFixation(skuSpec.getLength(), skuSpec.getWidth());

            if (fixation) {
                fixationList.add(orderSku);
            } else {
                notFixationList.add(orderSku);
            }
        }
        // 固定排版
        fixationScheduling(bomSpecId, printingPaperBomSpecId, fixationList, skuSpecMap);

        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(2);
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);

        // 非固定排版
        notFixationScheduling(workOrder, notFixationList, skuSpecMap, true);
    }

    /**
     * 固定排版
     */
    private void fixationScheduling(Long bomSpecId, Long printingPaperBomSpecId,
                                    List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap) {

        if (ObjectUtil.isEmpty(list)) {
            return;
        }

        // 根据固定排版规则区分
        Map<WorkOrderFixationSpecEnum, List<OrderSku>> map = list.stream().collect(Collectors.groupingBy(
                item -> {
                    SkuSpec skuSpec = getAndVerifySkuSpec(item.getSkuSpecId(), skuSpecMap);
                    return WorkOrderFixationSpecEnum.getEnum(skuSpec.getLength(), skuSpec.getWidth());
                }
        ));

        map.forEach(((k, v) -> doFixationScheduling(bomSpecId, printingPaperBomSpecId, k, v)));

    }

    /**
     * 固定排版
     */
    private void doFixationScheduling(Long bomSpecId, Long printingPaperBomSpecId,
                                      WorkOrderFixationSpecEnum workOrderFixationSpecEnum, List<OrderSku> list) {

        // 本次所需排版数量
        int schedulingNum = workOrderFixationSpecEnum.getMaxNum();

        // 未排单数量
        int notSchedulingNum = list.stream().map(item -> item.getQuantity().subtract(item.getSchedulingNum()))
                .reduce(BigDecimal.ZERO, BigDecimal::add).intValue();

        // 如果未排版数量小于本次所需排版数量，则不排单
        if (notSchedulingNum < schedulingNum) {
            return;
        }

        FixationStrategy fixationStrategy = FixationFactory.get(workOrderFixationSpecEnum);

        // 生产工单
        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(1);
        workOrder.setFixationSpec(workOrderFixationSpecEnum.getLength() + "*" + workOrderFixationSpecEnum.getWidth());
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);
        workOrder.setSyncProduction(StatusConstant.NO);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = new ArrayList<>();

        for (OrderSku orderSku : list) {

            // 此单未排版数量
            int itemSchedulingNum = orderSku.getQuantity().subtract(orderSku.getSchedulingNum()).intValue();

            // 未排排数量小于等于本次所需排版数量
            if (itemSchedulingNum <= schedulingNum) {

                // 赋值排版数量等于sku数量
                orderSku.setSchedulingNum(orderSku.getQuantity());
                orderSku.setCompleteScheduling(StatusConstant.YES);

                for (int i = 0; i < itemSchedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                // 重新计算需要排版数量
                schedulingNum -= itemSchedulingNum;
            }

            // 未排排数量大于本次所需排版数量
            else {
                orderSku.setSchedulingNum(orderSku.getSchedulingNum().add(new BigDecimal(schedulingNum)));
                orderSku.setCompleteScheduling(StatusConstant.NO);

                for (int i = 0; i < schedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                break;
            }

        }

        workOrder.setSchedulingNum(fixationStrategy.getSchedulingNum());
        workOrder.setUsedArea(fixationStrategy.getUsedArea());
        workOrder.setUseRatio(fixationStrategy.getUseRatio());
        workOrder.setMaxLength(fixationStrategy.getMaxLength());

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            save(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

        // 再次执行判断是否还可以排版
        doFixationScheduling(bomSpecId, printingPaperBomSpecId, workOrderFixationSpecEnum, list);
    }

    /**
     * 添加排版明细
     */
    private void addWorkOrderDetail(FixationStrategy fixationStrategy, WorkOrder workOrder,
                                    List<WorkOrderDetail> workOrderDetailList, OrderSku orderSku) {
        CoordinateBo coordinateBo = fixationStrategy.next();
        WorkOrderDetail workOrderDetail = new WorkOrderDetail();
        workOrderDetail.setWorkOrderId(workOrder.getId());
        workOrderDetail.setOrderInfoId(orderSku.getOrderId());
        workOrderDetail.setOrderSkuId(orderSku.getId());
        workOrderDetail.setLength(coordinateBo.getLength());
        workOrderDetail.setWidth(coordinateBo.getWidth());
        workOrderDetail.setBleedingLength(coordinateBo.getBleedingLength());
        workOrderDetail.setBleedingWidth(coordinateBo.getBleedingWidth());
        workOrderDetail.setRotate(coordinateBo.getHasRotate() ? StatusConstant.YES : StatusConstant.NO);
        workOrderDetail.setX(coordinateBo.getX());
        workOrderDetail.setY(coordinateBo.getY());
        workOrderDetailList.add(workOrderDetail);
    }

    /**
     * 非固定排版
     */
    private void notFixationScheduling(WorkOrder workOrder, List<OrderSku> list,
                                       Map<Long, SkuSpec> skuSpecMap, boolean carpeting) {

        if (ObjectUtil.isEmpty(list)) {
            return;
        }

        Item[] items = list.stream().flatMap(item -> {
            SkuSpec skuSpec = getAndVerifySkuSpec(item.getSkuSpecId(), skuSpecMap);
            int schedulingNum = item.getQuantity().subtract(item.getSchedulingNum()).intValue();

            List<Item> itemList = new ArrayList<>();
            for (int i = 0; i < schedulingNum; i++) {
                Long orderSkuId = item.getId();
                double length = skuSpec.getLength().doubleValue() + MaterialsConstant.RESERVE;
                double width = skuSpec.getWidth().doubleValue() + MaterialsConstant.RESERVE;

                itemList.add(new Item(orderSkuId, length, width));
            }

            return itemList.stream();
        }).toArray(Item[]::new);

        if (items.length > 430) {
            items = ArrayUtil.resize(items, 430);
        }

        Solution solve = GA.solve(MaterialsConstant.MASTER_LENGTH, MaterialsConstant.MASTER_WIDTH, items);

        List<PlaceItem> placeItemList = solve.getPlaceItemList();
        if (carpeting && placeItemList.size() == items.length) {
            return;
        }

        Map<Long, OrderSku> map = list.stream().collect(Collectors.toMap(BaseIdPo::getId, item -> item));

        // 生产工单
        workOrder.setSchedulingNum(new BigDecimal(placeItemList.size()));
        workOrder.setUsedArea(BigDecimal.valueOf(solve.getTotal()));
        workOrder.setUseRatio(BigDecimal.valueOf(solve.getUseRatio() * 100));
        workOrder.setMaxLength(BigDecimal.valueOf(solve.getMaxLength()));
        workOrder.setSyncProduction(StatusConstant.NO);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = placeItemList.stream().map(item -> {
            OrderSku orderSku = map.get(item.getId());

            BigDecimal schedulingNum = orderSku.getSchedulingNum().add(BigDecimal.ONE);
            orderSku.setSchedulingNum(schedulingNum);
            if (schedulingNum.compareTo(orderSku.getQuantity()) == 0) {
                orderSku.setCompleteScheduling(StatusConstant.YES);
            }

            WorkOrderDetail workOrderDetail = new WorkOrderDetail();
            workOrderDetail.setWorkOrderId(workOrder.getId());
            workOrderDetail.setOrderInfoId(orderSku.getOrderId());
            workOrderDetail.setOrderSkuId(orderSku.getId());
            workOrderDetail.setX(BigDecimal.valueOf(item.getX()));
            workOrderDetail.setY(BigDecimal.valueOf(item.getY()));
            workOrderDetail.setLength(BigDecimal.valueOf(item.getLength() - MaterialsConstant.RESERVE));
            workOrderDetail.setWidth(BigDecimal.valueOf(item.getWidth() - MaterialsConstant.RESERVE));
            workOrderDetail.setBleedingLength(BigDecimal.valueOf(item.getLength()));
            workOrderDetail.setBleedingWidth(BigDecimal.valueOf(item.getWidth()));
            workOrderDetail.setRotate(item.isRotate() ? StatusConstant.YES : StatusConstant.NO);
            return workOrderDetail;
        }).collect(Collectors.toList());

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            saveOrUpdate(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

    }

    /**
     * 公共排版方法
     */
    private void commonScheduling(CommonSchedulingFunInterface commonSchedulingFunInterface) {
        boolean b = schedulingLock.tryLock();

        if (!b) {
            throw new ServiceException("正在排版中，请稍后刷新数据");
        }

        try {

            // 查找未排版完成的订单sku
            List<OrderSku> list = getNotSchedulingSku(null, null, null, null);

            if (list.isEmpty()) {
                return;
            }

            // 获取需要排版的sku规格
            List<Long> skuSpecIdList = list.stream().map(OrderSku::getSkuSpecId).collect(Collectors.toList());
            Map<Long, SkuSpec> skuSpecMap = skuSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, skuSpecIdList));

            // 根据主材和打印纸区分订单sku
            HashBasedTable<Long, Long, List<OrderSku>> map = HashBasedTable.create();
            for (OrderSku orderSku : list) {
                List<OrderSku> itemList = map.get(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId());
                if (itemList == null) {
                    itemList = new ArrayList<>();
                    map.put(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId(), itemList);
                }
                itemList.add(orderSku);
            }

            // 排版
            for (Long bomSpecId : map.rowKeySet()) {
                for (Long printingPaperBomSpecId : map.columnKeySet()) {
                    List<OrderSku> orderSkuList = map.get(bomSpecId, printingPaperBomSpecId);
                    commonSchedulingFunInterface.run(bomSpecId, printingPaperBomSpecId, orderSkuList, skuSpecMap);
                }
            }

        } catch (ServiceException e) {
            throw e;
        } catch (Exception e) {
            log.error("排版业务发生异常", e);
            throw new ServiceException("排版业务发生异常");
        } finally {
            schedulingLock.unlock();
        }

    }

    /**
     * 获取并验证sku规格
     */
    private SkuSpec getAndVerifySkuSpec(Long skuSpecId, Map<Long, SkuSpec> skuSpecMap) {
        SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
        Assert.notNull(skuSpec, "未知sku规格，id：" + skuSpecId);
        Assert.notNull(skuSpec.getLength(), "品号为【" + skuSpec.getCode() + "】的sku未定义长度");
        Assert.notNull(skuSpec.getWidth(), "品号为【" + skuSpec.getCode() + "】的sku未定义宽度");
        return skuSpec;
    }

    /**
     * 获取未排版订单sku
     */
    private List<OrderSku> getNotSchedulingSku(Long bomSpecId, Long printingPaperBomSpecId, BigDecimal length, BigDecimal width) {
        return Sql.create(OrderSku.class)
                .selectAll(OrderSku.class)
                .from(OrderInfo.class)
                .innerJoin(OrderSku.class, OrderSku::getOrderId, OrderInfo::getId)
                .leftJoin(SkuSpec.class, OrderSku::getSkuSpecId)
                .eq(OrderInfo::getExceptionType, OrderExceptionTypeEnum.NORMAL.getKey())
                .eq(OrderSku::getCompleteScheduling, StatusConstant.NO)
                .eq(OrderSku::getBomSpecId, bomSpecId)
                .eq(OrderSku::getPrintingPaperBomSpecId, printingPaperBomSpecId)
                .eq(SkuSpec::getLength, length)
                .eq(SkuSpec::getWidth, width)
                .list();
    }

    /**
     * 获取公海工单
     */
    private List<OpenSeaWorkOrderVo> getOpenSeaWorkOrderList(List<OrderSku> orderSkuList) {
        List<OpenSeaWorkOrderVo> openSeaWorkOrderList = new ArrayList<>();
        for (OrderSku orderSku : orderSkuList) {
            int number = orderSku.getQuantity().subtract(orderSku.getSchedulingNum()).intValue();
            for (int i = 0; i < number; i++) {
                OpenSeaWorkOrderVo openSeaWorkOrderVo = new OpenSeaWorkOrderVo();
                openSeaWorkOrderVo.setOrderInfoId(orderSku.getOrderId());
                openSeaWorkOrderVo.setOrderSkuId(orderSku.getId());
                openSeaWorkOrderVo.setSkuSpecId(orderSku.getSkuSpecId());
                openSeaWorkOrderList.add(openSeaWorkOrderVo);
            }
        }
        return openSeaWorkOrderList;
    }

    /**
     * 原料出库
     */
    private void ingredientsOut(WorkOrder workOrder) {
        InOutStorageDto inOutStorageDto = new InOutStorageDto();
        inOutStorageDto.setType(InOutTypeEnum.OUT.getKey());
        inOutStorageDto.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        inOutStorageDto.setClassifyType(1);
        inOutStorageDto.setWarehouseId(WarehouseConstant.SEMI_FINISHED_PRODUCT);
        inOutStorageDto.setRemark("生产工单【" + workOrder.getCode() + "】原料生产出库");
        List<InOutStorageBom> inOutStorageBomList = new ArrayList<>();
        inOutStorageDto.setInOutStorageBomList(inOutStorageBomList);

        // 转印布
        InOutStorageBom master = new InOutStorageBom();
        master.setBomSpecId(workOrder.getMasterBomSpecId());
        master.setQuantity(BigDecimal.ONE);
        master.setMeter(workOrder.getMasterLength().divide(new BigDecimal("100")));
        inOutStorageBomList.add(master);

        // 打印纸
        InOutStorageBom printingPaper = new InOutStorageBom();
        printingPaper.setBomSpecId(workOrder.getPrintingPaperBomSpecId());
        printingPaper.setQuantity(BigDecimal.ONE);
        printingPaper.setMeter(workOrder.getPrintingPaperLength().divide(new BigDecimal("100")));
        inOutStorageBomList.add(printingPaper);

        inStorageService.add(inOutStorageDto);
    }

    /**
     * 余料入库
     */
    private void surplusIn(WorkOrder workOrder) {

        List<InOutStorageBom> inOutStorageBomList = new ArrayList<>();

        // 转印布
        BigDecimal masterMeter = workOrder.getMasterLength().subtract(workOrder.getMaxLength()).divide(new BigDecimal("100"));
        if (masterMeter.compareTo(new BigDecimal("0.4")) >= 0) {
            InOutStorageBom master = new InOutStorageBom();
            master.setBomSpecId(workOrder.getMasterBomSpecId());
            master.setQuantity(BigDecimal.ONE);
            master.setMeter(masterMeter);
            inOutStorageBomList.add(master);
        }

        // 打印纸
        BigDecimal printingPaperMeter = workOrder.getPrintingPaperLength().subtract(workOrder.getMaxLength()).divide(new BigDecimal("100"));
        if (printingPaperMeter.compareTo(new BigDecimal("0.4")) >= 0) {
            InOutStorageBom printingPaper = new InOutStorageBom();
            printingPaper.setBomSpecId(workOrder.getPrintingPaperBomSpecId());
            printingPaper.setQuantity(BigDecimal.ONE);
            printingPaper.setMeter(printingPaperMeter);
            inOutStorageBomList.add(printingPaper);
        }

        if (!inOutStorageBomList.isEmpty()) {
            InOutStorageDto inOutStorageDto = new InOutStorageDto();
            inOutStorageDto.setType(InOutTypeEnum.IN.getKey());
            inOutStorageDto.setDetailType(InDetailTypeEnum.RETURN_GOODS.getKey());
            inOutStorageDto.setClassifyType(1);
            inOutStorageDto.setWarehouseId(WarehouseConstant.SEMI_FINISHED_PRODUCT);
            inOutStorageDto.setRemark("生产工单【" + workOrder.getCode() + "】余料退料入库");
            inOutStorageDto.setInOutStorageBomList(inOutStorageBomList);
            inStorageService.add(inOutStorageDto);
        }

    }

    /**
     * 包材出库
     */
    private void packagingMaterialOut(WorkOrder workOrder, List<WorkOrderDetail> list) {
        InOutStorageDto inOutStorageDto = new InOutStorageDto();
        inOutStorageDto.setType(InOutTypeEnum.OUT.getKey());
        inOutStorageDto.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        inOutStorageDto.setClassifyType(2);
        inOutStorageDto.setWarehouseId(WarehouseConstant.PACKAGING_MATERIAL);
        inOutStorageDto.setRemark("生产工单【" + workOrder.getCode() + "】包材生产出库");

        Map<Long, InOutStorageBom> map = new HashMap<>();

        // 通过工单找到订单sku包材配置
        List<Long> orderSkuIdList = list.stream().map(WorkOrderDetail::getOrderSkuId).collect(Collectors.toList());
        Map<Long, List<OrderSkuBom>> orderSkuMap = orderSkuBomService.mapKGroup(OrderSkuBom::getOrderSkuId,
                q -> q.in(OrderSkuBom::getOrderSkuId, orderSkuIdList));

        // 循环订单sku id
        for (Long orderSkuId : orderSkuIdList) {

            // 获取订单skuId的包材
            List<OrderSkuBom> orderSkuBomList = orderSkuMap.getOrDefault(orderSkuId, Collections.emptyList());

            // 循环包材，添加入库数量
            for (OrderSkuBom orderSkuBom : orderSkuBomList) {
                InOutStorageBom inOutStorageBom = map.computeIfAbsent(orderSkuBom.getBomSpecId(), bomSpecId -> {
                    InOutStorageBom tempInOutStorageBom = new InOutStorageBom();
                    tempInOutStorageBom.setBomSpecId(bomSpecId);
                    tempInOutStorageBom.setQuantity(BigDecimal.ZERO);
                    return tempInOutStorageBom;
                });
                inOutStorageBom.setQuantity(inOutStorageBom.getQuantity().add(orderSkuBom.getQuantity()));
            }
        }
        inOutStorageDto.setInOutStorageBomList(new ArrayList<>(map.values()));

        // 执行入库操作
        inStorageService.add(inOutStorageDto);
    }


    /**
     * 重新排单
     */
    private void reschedule(WorkOrder workOrder, List<WorkOrderDetail> workOrderDetailList) {

        // 固定尺寸排版
        if (Objects.equals(workOrder.getType(), 1)) {

            String[] split = workOrder.getFixationSpec().split("\\*");
            BigDecimal length = new BigDecimal(split[0]);
            BigDecimal width = new BigDecimal(split[1]);

            WorkOrderFixationSpecEnum workOrderFixationSpecEnum = WorkOrderFixationSpecEnum.getEnum(length, width);

            FixationStrategy fixationStrategy = FixationFactory.get(workOrderFixationSpecEnum);

            for (WorkOrderDetail orderDetail : workOrderDetailList) {
                CoordinateBo coordinateBo = fixationStrategy.next();

                orderDetail.setLength(coordinateBo.getLength());
                orderDetail.setWidth(coordinateBo.getWidth());
                orderDetail.setBleedingLength(coordinateBo.getBleedingLength());
                orderDetail.setBleedingWidth(coordinateBo.getBleedingWidth());
                orderDetail.setRotate(coordinateBo.getHasRotate() ? StatusConstant.YES : StatusConstant.NO);
                orderDetail.setX(coordinateBo.getX());
                orderDetail.setY(coordinateBo.getY());
            }

            List<OrderSku> notSchedulingSku = getNotSchedulingSku(workOrder.getMasterBomSpecId(),
                    workOrder.getPrintingPaperBomSpecId(), length, width);

            for (OrderSku orderSku : notSchedulingSku) {

                if (fixationStrategy.hasNext()) {

                    int number = orderSku.getQuantity().subtract(orderSku.getSchedulingNum()).intValue();
                    for (int i = 0; i < number; i++) {
                        if (fixationStrategy.hasNext()) {
                            addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                            orderSku.setSchedulingNum(orderSku.getSchedulingNum().add(BigDecimal.ONE));
                            orderSku.setCompleteScheduling(orderSku.getSchedulingNum()
                                    .compareTo(orderSku.getQuantity()) == 0 ? StatusConstant.YES : StatusConstant.NO);
                        } else {
                            break;
                        }
                    }

                } else {
                    break;
                }
            }

            workOrder.setSchedulingNum(fixationStrategy.getSchedulingNum());
            workOrder.setUsedArea(fixationStrategy.getUsedArea());
            workOrder.setUseRatio(fixationStrategy.getUseRatio());
            workOrder.setMaxLength(fixationStrategy.getMaxLength());

            if (!notSchedulingSku.isEmpty()) {
                orderSkuService.updateBatchById(notSchedulingSku);
            }
            updateById(workOrder);
            workOrderDetailService.saveOrUpdateBatch(workOrderDetailList);
        }
        // 非固定尺寸排版
        else {
            List<Long> orderSkuIdList = workOrderDetailList.stream().map(WorkOrderDetail::getOrderSkuId).collect(Collectors.toList());
            Map<Long, OrderSku> map = new HashMap<>();
            if (!orderSkuIdList.isEmpty()) {
                map = orderSkuService.listByIds(orderSkuIdList).stream()
                        .collect(Collectors.toMap(BaseIdPo::getId, Function.identity(), (t1, t2) -> t1));
            }

            for (WorkOrderDetail workOrderDetail : workOrderDetailList) {
                OrderSku orderSku = map.get(workOrderDetail.getOrderSkuId());
                orderSku.setSchedulingNum(orderSku.getSchedulingNum().subtract(BigDecimal.ONE));
                orderSku.setCompleteScheduling(StatusConstant.NO);
            }
            orderSkuService.updateBatchById(map.values());
            workOrderDetailService.remove(q -> q.eq(WorkOrderDetail::getWorkOrderId, workOrder.getId()));

            List<OrderSku> notSchedulingSku = getNotSchedulingSku(workOrder.getMasterBomSpecId(),
                    workOrder.getPrintingPaperBomSpecId(), null, null);

            Set<Long> skuSpecIdSet = notSchedulingSku.stream().map(OrderSku::getSkuSpecId).collect(Collectors.toSet());
            Map<Long, SkuSpec> skuSpecMap = skuSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, skuSpecIdSet));
            notFixationScheduling(workOrder, notSchedulingSku, skuSpecMap, false);
        }
    }

    @FunctionalInterface
    private interface CommonSchedulingFunInterface {
        void run(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap);
    }

}
